from __future__ import absolute_import

import socket
from urllib3.connection import HTTPSConnection, HTTPConnection
from urllib3.util.connection import allowed_gai_family, _set_socket_options
from urllib3.exceptions import ConnectTimeoutError, NewConnectionError, LocationParseError
from urllib3.packages import six
from socket import error as SocketError
from socket import timeout as SocketTimeout

from .resolver import resolve_dns

# This code is copied from urllib3/util/connection.py version 1.26.8 (from requests v2.28.1)
def create_connection(
    address,
    timeout=socket._GLOBAL_DEFAULT_TIMEOUT,
    source_address=None,
    socket_options=None,
):
    """Same as :meth:`urllib3.util.connection.create_connection()`, 
    except it has DNS over HTTPS resovler inside of it.
    """

    host, port = address
    if host.startswith("["):
        host = host.strip("[]")
    err = None

    # Using the value from allowed_gai_family() in the context of getaddrinfo lets
    # us select whether to work with IPv4 DNS records, IPv6 records, or both.
    # The original create_connection function always returns all records.
    family = allowed_gai_family()

    try:
        host.encode("idna")
    except UnicodeError:
        return six.raise_from(
            LocationParseError(u"'%s', label empty or too long" % host), None
        )

    orig_answers = socket.getaddrinfo(host, port, family, socket.SOCK_STREAM)

    try:
        af, socktype, proto, canonname, sa = orig_answers[0]
    except IndexError:
        raise socket.error("getaddrinfo returns an empty list")

    answers = resolve_dns(host)

    for answer in answers:
        sa = (answer['data'], 443)
        sock = None
        try:
            sock = socket.socket(af, socktype, proto)

            # If provided, set socket level options before connecting.
            _set_socket_options(sock, socket_options)

            if timeout is not socket._GLOBAL_DEFAULT_TIMEOUT:
                sock.settimeout(timeout)
            if source_address:
                sock.bind(source_address)
            sock.connect(sa)
            return sock

        except socket.error as e:
            err = e
            if sock is not None:
                sock.close()
                sock = None

    if err is not None:
        raise err

class ModifiedHTTPConnection(HTTPConnection):
    # This code is copied from urllib3/connection.py version 1.26.8 (from requests v2.28.1)
    def _new_conn(self):
        """Establish a socket connection and set nodelay settings on it.

        :return: New socket connection.
        """
        extra_kw = {}
        if self.source_address:
            extra_kw["source_address"] = self.source_address

        if self.socket_options:
            extra_kw["socket_options"] = self.socket_options

        try:
            conn = create_connection(
                (self._dns_host, self.port), self.timeout, **extra_kw
            )

        except SocketTimeout:
            raise ConnectTimeoutError(
                self,
                "Connection to %s timed out. (connect timeout=%s)"
                % (self.host, self.timeout),
            )

        except SocketError as e:
            raise NewConnectionError(
                self, "Failed to establish a new connection: %s" % e
            )

        return conn

class ModifiedHTTPSConnection(ModifiedHTTPConnection, HTTPSConnection):
    pass



from requests.adapters import HTTPAdapter
from urllib3.connectionpool import HTTPSConnectionPool, HTTPConnectionPool

from .connector import ModifiedHTTPConnection, ModifiedHTTPSConnection
from .resolver import set_dns_provider

__all__ = ('DNSOverHTTPSAdapter',)

class DNSOverHTTPSAdapter(HTTPAdapter):
    """An DoH (DNS over HTTPS) adapter for :class:`requests.Session`
    
    Parameters
    -----------
    provider: :class:`str`
        A DoH provider
    **kwargs
        These parameters will be passed to :meth:`requests.adapters.HTTPAdapter.__init__()`
    """
    def __init__(self, provider=None, **kwargs):
        if provider:
            set_dns_provider(provider)

        super().__init__(**kwargs)

    def get_connection(self, url, proxies=None):
        conn = super().get_connection(url, proxies)
        if isinstance(conn, HTTPSConnectionPool):
            conn.ConnectionCls = ModifiedHTTPSConnection
        else:
            # HTTP type
            conn.ConnectionCls = ModifiedHTTPConnection
        return conn


import pathlib
import re
from setuptools import setup, find_packages

# Root directory
# (README.md, requests_doh/__init__.py)
HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()
init_file = (HERE / "requests_doh/__init__.py").read_text()

def get_version():
    """Get version of the app"""
    re_version = r'__version__ = \"([0-9]{1,3}.[0-9]{1,3}.[0-9]{1,3})\"'
    _version = re.search(re_version, init_file)

    if _version is None:
        raise RuntimeError("Version is not set")

    return _version.group(1)

def get_value_var(var_name):
    """Get value of `__{var_name}__` from `requests_doh/__init__.py`"""
    var = f'__{var_name}__'
    regex = '%s = "(.{1,})"' % var

    found = re.search(regex, init_file)

    if found is None:
        raise RuntimeError(f'{var} is not set in "requests_doh/__init__.py"')
    
    return found.group(1)

def get_requirements():
    """Return tuple of library needed for app to run"""
    main = []
    try:
        with open('./requirements.txt', 'r') as r:
            main = r.read().splitlines()
    except FileNotFoundError:
        raise RuntimeError("requirements.txt is needed to build requests-doh")

    if not main:
        raise RuntimeError("requirements.txt have no necessary libraries inside of it")

    docs = []
    try:
        with open('./requirements-docs.txt', 'r') as r:
            docs = r.read().splitlines()
    except FileNotFoundError:
        # There is no docs requirements
        # Developers can ignore this error and continue to install without any problem.
        # However, this is needed if developers want to create documentation in readthedocs.org or local device.
        pass

    return main, {
        "docs": docs,
    }

# Get requirements needed to build app
requires_main, extras_require = get_requirements()

# Get all modules
packages = find_packages('.')

# Get repository
repo = get_value_var('repository')

# Finally run main setup
setup(
    name = 'requests-doh',         
    packages = packages,   
    version = get_version(),
    license=get_value_var('license'),
    description = get_value_var('description'),
    long_description= README,
    long_description_content_type= 'text/markdown',
    author = get_value_var('author'),
    author_email = get_value_var('author_email'),
    url = f'https://github.com/{repo}',
    keywords = [
        'requests',
        'doh',
        'dns',
        'https',
        'dns-over-https'
    ], 
    install_requires=requires_main,
    classifiers=[
        'Development Status :: 5 - Production/Stable',  
        'Intended Audience :: Developers',
        'Topic :: Internet :: Name Service (DNS)',
        'Topic :: Internet :: WWW/HTTP',
        'License :: OSI Approved :: MIT License',  
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10'
    ],
    python_requires='>=3.8'
)
